import { Effect, Option, Sink, Stream } from "effect";
import { RawAccount } from "cojson";
import { Account, AccountCtx, Group, SubscriptionScope, Ref, inspect, subscriptionsScopes, } from "../internal.js";
import { fulfillsDepth } from "./deepLoading.js";
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export function isCoValue(value) {
    return value && value._type !== undefined;
}
export function isCoValueClass(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
value) {
    return typeof value === "function" && value.fromRaw !== undefined;
}
/** @internal */
export class CoValueBase {
    get _owner() {
        const owner = this._raw.group instanceof RawAccount
            ? Account.fromRaw(this._raw.group)
            : Group.fromRaw(this._raw.group);
        const subScope = subscriptionsScopes.get(this);
        if (subScope) {
            subScope.onRefAccessedOrSet(this.id, owner.id);
            subscriptionsScopes.set(owner, subScope);
        }
        return owner;
    }
    /** @private */
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    constructor(..._args) {
        Object.defineProperty(this, "_instanceID", {
            value: `instance-${Math.random().toString(36).slice(2)}`,
            enumerable: false,
        });
    }
    /** @category Internals */
    static fromRaw(raw) {
        return new this({ fromRaw: raw });
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    toJSON() {
        return {
            id: this.id,
            type: this._type,
            error: "unknown CoValue class",
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    /** @category Type Helpers */
    castAs(cl) {
        return cl.fromRaw(this._raw);
    }
}
export function loadCoValue(cls, id, as, depth) {
    return Effect.runPromise(loadCoValueEf(cls, id, depth).pipe(Effect.mapError(() => undefined), Effect.merge, Effect.provideService(AccountCtx, as)));
}
export function ensureCoValueLoaded(existing, depth) {
    return loadCoValue(existing.constructor, existing.id, existing._loadedAs, depth);
}
export function loadCoValueEf(cls, id, depth) {
    return subscribeToCoValueEf(cls, id, depth).pipe(Stream.runHead, Effect.andThen(Effect.mapError((_noSuchElem) => "unavailable")));
}
export function subscribeToCoValue(cls, id, as, depth, listener) {
    void Effect.runPromise(Effect.provideService(subscribeToCoValueEf(cls, id, depth).pipe(Stream.run(Sink.forEach((update) => Effect.sync(() => listener(update))))), AccountCtx, as));
    return function unsubscribe() { };
}
export function subscribeToExistingCoValue(existing, depth, listener) {
    return subscribeToCoValue(existing.constructor, existing.id, existing._loadedAs, depth, listener);
}
export function subscribeToCoValueEf(cls, id, depth) {
    return AccountCtx.pipe(Effect.andThen((account) => new Ref(id, account, {
        ref: cls,
        optional: false,
    }).loadEf()), Stream.fromEffect, Stream.flatMap((value) => Stream.asyncScoped((emit) => Effect.gen(function* (_) {
        const subscription = new SubscriptionScope(value, cls, (update) => void emit.single(update));
        yield* _(Effect.addFinalizer(() => Effect.sync(() => subscription.unsubscribeAll())));
    }))), Stream.filterMap((update) => Option.fromNullable(fulfillsDepth(depth, update)
        ? update
        : undefined)));
}
//# sourceMappingURL=interfaces.js.map